// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****
//
// 
// ForkOperations implements most of the fork-window operations, like
// shadowing the main window and turning cookies on or off.
//
// It should be instantiated immediately after creating the window, as
// it relies on an onLoad handler being fired.
//
// forkBrowserOverlay.js just does window management tasks like
// handling setup and resize requests.
//

var NS_BINDING_ABORTED=0x804b0002;
var NS_BINDING_FAILED=0x804b0001;
var NS_BINDING_SUCCEEDED=0;

var _NEXT_ID = 1;
function getNextID() {
    return _NEXT_ID++;
};

function loadVisitCounts() {
    var prefObj = getPrefService();
    var Branch = prefObj.getBranch("cookielog.");

    if ( ! Branch.prefHasUserValue("visitcounts") ) {
        return new Array();
    }

    var countString = Branch.getCharPref("visitcounts");
    var tmp = countString.split(":");

    var counts = new Array();
    for (var i in tmp) {
        var tmp2 = tmp[i].split("=");
        counts[tmp2[0]] = tmp2[1];
    }

    return counts;
}

function storeVisitCounts(visitCounts) {
    var prefObj = getPrefService();
    var Branch = prefObj.getBranch("cookielog.");

    var countString = "";
    for (var domain in visitCounts) {
        if ( countString != "" ) {
            countString += ":";
        }

        countString += (domain + "=" + visitCounts[domain]);
    }

    Branch.setCharPref("visitcounts", countString);
}


// Some state needs to be shared between all ForkOperations instances
function InitForkGlobalState() {
    this.SITE_MAX_STEP_COUNT  = null;    // after this many page loads, assume no more cookies needed
    this.SITE_MAX_VISIT_COUNT = null;    // after this many visits to the site, assume no more cookies needed
    
    var prefObj = getPrefService();
    var Branch = prefObj.getBranch("cookielog.");

    if ( ! Branch.prefHasUserValue("MAX_STEP_COUNT") ) {
        Branch.setIntPref("MAX_STEP_COUNT", 5);
    }
    this.SITE_MAX_STEP_COUNT = Branch.getIntPref("MAX_STEP_COUNT");

    if ( ! Branch.prefHasUserValue("MAX_VISIT_COUNT") ) {
        Branch.setIntPref("MAX_VISIT_COUNT", 8);
    }
    this.SITE_MAX_VISIT_COUNT = Branch.getIntPref("MAX_VISIT_COUNT");

    this.doneDomains = new Array(); // on which domains did we already make a decision session (don't fork again)
                                    // 0/null = didn't; 1 = did w/positive perms; 2 = did w/negative perms
    
    this.visitCounts = loadVisitCounts();   // how many times have we visited each tailDomain
    this.unload = function() { storeVisitCounts(this.visitCounts) };
    
    this.visitSite = function(tailDomain) {
        if ( this.visitCounts[tailDomain] == null ) {
            this.visitCounts[tailDomain] = 1;
        } else {
            this.visitCounts[tailDomain]++;
        }

        dumpToConsole("Increasing visitCount for site " + tailDomain + " to " + this.visitCounts[tailDomain], "forkwindow-debug");

        // -1 = not there yet
        // 0 = just exceeded the limit
        // > 0 = past the limit
        return ( this.visitCounts[tailDomain] - this.SITE_MAX_VISIT_COUNT );
    }

    this.resetVisitCount = function (tailDomain) {
        dumpToConsole("Resetting visitCount for site " + tailDomain + " (due to the user making a choice) ");
        this.visitCounts[tailDomain] = 0;
    }

    this.clearVisitCounts = function() {
        this.visitCounts = new Array();
        this.doneDomains = new Array();
        storeVisitCounts(this.visitCounts);
    }
    
    return this;
};
 
var forkGlobalState = new InitForkGlobalState();
 
// Write the visit counts back when we quit
window.addEventListener("unload", function() { forkGlobalState.unload() }, true);

function ForkOperations(forkWin, mainDOMWin, cookielog, cookieperms) {
    //    this.ID = getNextID();
    this.win = forkWin;
    //    this.mainWin = mainWin;
    this.mainDOMWin = mainDOMWin; 
    this.domWindow = null;
    this.cookielog = cookielog;     // oCookieLog
    this.cookieperms = cookieperms; // oCookiePermissions    
    this.cookiePermissions = FP_SESSION;

    this._mainDone = true; // keep track of the loading states
    this._forkDone = true;
    this._doSync = false;
    this._loadPending = false;
    this._clickSinceLastStart = false;
    this._tailDomainRE= null;
    this._lastTailDomain = null;
    this._lastTailDomain = null;
    this._lastP3PHeader = null;
    this._statusWindow = null;

    this.sync = "";

    this.setSync = function (newSync) {
        this.sync = newSync;
        dumpToConsole("Setting fork sync to " + newSync, "forkwindow-debug");
        notifyObservers(this.getMainDOMWindow(), "set-statusbar-sync", newSync);
    }
    
    addObserver(this, "login-form-detected", false);
    
    var localThis = this;

    // The point of this is that focus()ing the fork window un-hides
    // it and we don't like that, most of the time
    this._killForkFocus = true;
    forkWin.addEventListener("focus",
                             function() {
        if ( ! cookielog.debug && ! localThis._killForkFocus ) {
            hideDOMOrChromeWindow(forkWin);
        }
    }, true);
    
    this.killForkFocus = function() {
        this._killForkFocus = true;
    }

    this.noKillForkFocus = function() {
        this._killForkFocus = false;
    }

    this.getForkWindow = function() { return this.win; };
    
    this.getAppContent = function() { 
        return this.appContent || (this.appContent = appContent(this.win)); 
    };

    this.getDOMWindow = function() {
        var win = (this.domWindow || (this.domWindow = htmlWinFromXULWindow(this.win)));
        return win;
    };

//     this.getMainWindow = function() {
//         return this.mainWin;
//     };

    this.getMainDOMWindow = function() {
        return this.mainDOMWin;
    };

    this.isContained = function(domWindow) {
        var thisDOM = this.getDOMWindow();
        return (domWindow == thisDOM ||
                domWindow.top == thisDOM || 
                (domWindow.opener && domWindow.opener.top == thisDOM));	       
    };

    //
    // Clear cookies only accepted because of this window (not the main one)
    //    

    // Default is to clear those for the current tailDomain
    this.clearAcceptedCookies = function(clearAll, tailDomain) {
        var cookieManager = getCookieManager();
        var cookies = cookieManager.enumerator;
        var prefix = "FORK"; //+ this.ID;
        var RE = ( tailDomain ? ( new RegExp(tailDomain + "$", "i") ) :
                   ( this._tailDomainRE ) );
        
        while (cookies.hasMoreElements()) {
            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie2);
            if (cookie.name.substr(0, prefix.length) == prefix) {
                if ( clearAll || cookie.host.match(RE) ) {
                    cookieManager.remove(cookie.host, cookie.name, cookie.path, false);
                }
            }
        }
    };

    //
    // Mirroring clicks is the main mechanism used to keep the fork
    // window in sync When an "out of band" event happens, e.g.,
    // something from chrome, that's synced using the load() method,
    // which directly loads the URL.
    //
    // After a certain number (SITE_MAX_STEP_COUNT) of page loads on a site,
    // we stop mirroring.
    //
    this.stepCount = 0;
    this.stillMirroring = function() {
        return this.stepCount < forkGlobalState.SITE_MAX_STEP_COUNT;
    }
    
    // msg gives a string reason why mirroring was stopped
    this.stopMirroring = function(msg) {
        this.setSync("NOFORK");
        this.stepCount = forkGlobalState.SITE_MAX_STEP_COUNT + 1;
        this.overlayStatusMessage(msg);
        this._waitingFor = null;
        dumpToConsole("stopMirroring() called with message " + msg);
    }

    this.sendClick = function(cookielog, clickIdx) {
        if ( ! this.stillMirroring() ) {
            return true;
        }

        this._forkDone = false;

        this._clickSinceLastStart = true;

        try {
            this.cookielog.showElement(this.getForkWindow(), clickIdx, false, true, false);
        }
        catch (e) {
            if ( String(e).match(/Could not find/) ) {
                dumpToConsole("Could not mirror the click in the fork window.", "forkwindow-debug");
                return false;
            }            
        }

        this.lastClickIdx = clickIdx;

        return true;
    };

    this._isFrameset = function(win) {
        return ( win.document.getElementsByTagName("FRAMESET").length );
        
    }

    // _compareWindows returns this._dummyMatched if the windows were the
    // same, or the fork window/frame that was different otherwise
    this._dummyMatched = new Object();
    this._compareWindows = function(mainDOM, forkDOM) {
        if (mainDOM.document.body &&
            forkDOM.document.body &&
            ! mainDOM.document.body.textContent.match(this.cookielog.userNameRE) &&
            forkDOM.document.body.textContent.match(this.cookielog.userNameRE) ) {
            return forkDOM;
        }

        // Sadly, a different number of frames is not grounds to declare a difference
        // Amazon has a new frame for with-cookie pages, but it's an empty javascript URL.
        // Better ideas welcome...
        if (mainDOM.frames.length == forkDOM.frames.length) {        
            for (var i = 0; i < mainDOM.frames.length; i++) {
                var ret = this._compareWindows(mainDOM.frames[i], forkDOM.frames[i]);
                if ( ret != this._dummyMatched ) {
                    return ret;
                }
            }
        }
            
        return this._dummyMatched;
    };

    this.compareAgainstMain = function() {
        var forkDOM = this.getDOMWindow();
        var mainDOM = this.getMainDOMWindow();
        
        var overlayWin = ( this._isFrameset(forkDOM) ) ? forkDOM.frames[0] : forkDOM; 


        if ( mainDOM.document.title != forkDOM.document.title ) {

            var doc = overlayWin.document;
            if ( doc.body ) {
                var makeoverlay = function() {
                    var doc = overlayWin.document;
                    if ( doc.getElementById("diffcallout") ) {
                        return;
                    }

                    var overlay = doc.createElement("div");
                    overlay.id = "diffcallout";
                    overlay.setAttribute( "style",
                                          "position: absolute; top: 0px; left: 0px; " + 
                                          "background: #CCFFCC; "+
                                          "-moz-border-radius:20px; " + 
                                          "font-family: sans-serif; font-style: normal; font-size: 12pt; font-weight: normal; " + 
                                          "border: solid black 2px; width: 270px; height: 120px; padding: 5px; z-index: 999'"
                                          ); 
                    overlay.innerHTML = "<img style='float: left; margin-right: 5px; margin-top: 20px; margin-bottom: 20px' width='34' height='30' src='http://upload.wikimedia.org/wikipedia/commons/5/5d/Up-1.png'>" + 
                    "The title of this page is different, indicating either that (1) the site may requires cookies to operate or (2) It has an auto-login or personalization feature.";                
                    doc.body.appendChild(overlay);
                    forkDOM.scrollTo(0,0);
                    mainDOM.scrollTo(0,0);
                }

                makeoverlay();
                setTimeout(makeoverlay, 3000);
            }                    

            dumpToConsole("Document titles different between fork (" + forkDOM.document.title +
                          ")  and main (" + mainDOM.document.title + ")", "forkwindow-debug");
            return false;
        }

        var matched = this._compareWindows(mainDOM, forkDOM);
        if ( matched != this._dummyMatched ) {

            var makeoverlay = function() {
                var doc = matched.document;
                if ( doc.getElementById("diffhilite") ) {
                    return;
                }
                var IH = doc.body.innerHTML;
                doc.body.innerHTML = 
                IH.replace(cookielog.userNameRE, 
                           "<span id='diffhilite' style='color: red; font-size: larger; background-color: yellow'>" + 
                           "$&</span>");

                var placeholder = doc.getElementById("placeholder");
                var newspan1 = doc.getElementById("diffhilite");
                var coords = getDocOffset(newspan1);
                
                var overlay = doc.createElement("div");
                overlay.id = "diffcallout";
                overlay.setAttribute( "style",
                                      "position: absolute; top: " + (coords["y"] + newspan1.offsetHeight) + "px; " +
                                      "left: " + (coords["x"]) + "px; " + 
                                      "background: #CCFFCC; "+
                                      "-moz-border-radius:20px; " + 
                                      "font-family: sans-serif; font-style: normal; font-size: 12pt; font-weight: normal; " + 
                                      "border: solid black 2px; width: 270px; height: 72px; padding: 5px; padding-right: 10px; padding-top: 10px; z-index: 999'"
                                      ); 
                overlay.innerHTML = "<img style='float: left; margin-right: 5px; margin-top: 20px; margin-bottom: 20px' width='34' height='30' src='http://upload.wikimedia.org/wikipedia/commons/5/5d/Up-1.png'/>" +          
                "Your name or login ID is present on this page, indicating a personalization feature.</div>";
                doc.body.appendChild(overlay);
                forkDOM.scrollTo(coords["x"], 0);
                newspan1.scrollIntoView(); 
                
                dumpToConsole("scrolling main to " + forkDOM.scrollX + "," + forkDOM.scrollY);
                mainDOM.scrollTo(0,0);                
                mainDOM.scrollBy(forkDOM.scrollX, forkDOM.scrollY);
            }
            makeoverlay();
            setTimeout(makeoverlay, 3000);

            dumpToConsole("User name/email matched in fork window but not in main", "forkwindow-debug");

            return false;

        }
        
        return true;
    };

    this.recordPermissions = function(url) {
        var tailDomain = getTailURLDomain(getURLDomain(url));
        forkGlobalState.doneDomains[tailDomain] = 1;
        cookieperms.setDomainCookiePerms( tailDomain, this.cookiePermissions );        
    };


    // Record that we *don't* want any cookies not already permitted in the main window
    this.recordNegativePermissions = function(url) {
        forkGlobalState.doneDomains[tailDomain] = 2;
        var tailDomain = getTailURLDomain(getURLDomain(url));
        var intMainPerms = cookieperms.getPermissions( this.getMainDOMWindow(),
                                                       tailDomain );
        var finalPerms = 0;
        for (var i = 1; i <= MAX_POS_PERM; i <<= 1) {
            if ( intMainPerms & i ) {
                finalPerms |= i;
            } else {
                finalPerms |= (i << POS_TO_NEG_PERMS);
            }
        }

        dumpToConsole("Setting new cookie permissions for " + tailDomain + ": " + parseCookiePermissionsToString(finalPerms));
        cookieperms.setDomainCookiePerms( tailDomain, finalPerms );
    };

    // both arguments optional; defaults: downgrade=false and taildomain=(all domains)
    this.copyCookiesToMain = function(doDowngrade, tailDomain) {        
        var cookieManager = getCookieManager();
        var cookies = cookieManager.enumerator;
        var prefix = "FORK"; // + this.ID;
        while (cookies.hasMoreElements()) {
            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie);
            cookie.QueryInterface(Components.interfaces.nsICookie2);

            if (cookie.name.substr(0, prefix.length) == prefix) {
                var newCookie = new Cookie(cookie);

                // FIXME: Need real domain matching
                if ( newCookie.domain.match(this._tailDomainRE) >= 0 ) {
                    continue;
                }

                newCookie.name = newCookie.name.substr(prefix.length);
                if (doDowngrade) {
                    newCookie.expires = null;
                    newCookie.isSession = true;
                }
                
                if ( ! newCookie.add() ) {
                    debugger;
                    throw ("Problem adding cookie in copyCookiesToMain()");
                }
            }
        }
    };

    // If we decide that cookies are needed for real, we need to turn on the perms for the main window
    // FIXME: Implement fine-grained comparison

    // NOTE: This function opens a non-modal window that serves a 
    // modal-like purpose, so don't do anything interesting after calling this function!
    this.suspectDifference = function() {
        if ( this._doSync ) {
            // We've been down this road already on a previous click, 
            // so just let the sync happen later
            return;
        }
        
        dumpToConsole("suspectDifference() called: auto-enabling cookies/side-by-side comparing", "forkwindow-debug");

        var url = canonicalizeURL(this.getDOMWindow().location.href);
        var tailDomain = getTailURLDomain(getURLDomain(url));
        var forkOp = this;

        var syncToMain = function() {
            forkGlobalState.resetVisitCount(tailDomain);
            forkOp.recordPermissions(url);
            forkOp.copyCookiesToMain(false, tailDomain);
            forkOp.clearAcceptedCookies(false,tailDomain);
            forkOp.stopMirroring("Difference detected; fork path chosen");
            forkOp.getMainDOMWindow().location = forkOp.getDOMWindow().location;
        }
        
        if ( ( (this.cookielog.paranoia == "paranoia-medium") || (this.cookielog.paranoia == "paranoia-low") )
             && this.cookiePermissions == FP_SESSION ) 
        {                 
            var doneButtonHTML = "<html:div style='margin-top: 10px; margin-left:35px'><button id='doneButton' label='Done!' oncommand='window.close()' style='visibility: hidden'/></html:div>"
            this._statusWindow = openStatusWindow(this.getMainDOMWindow(),
                                                  "Switching to cookie-enabled browser...", 
                                                  "Doppelganger has detected that cookies may be needed to use this site. " + 
                                                  "Please wait briefly while it enables cookies and reloads the page." +
                                                  doneButtonHTML,
                                                  this.getMainDOMWindow(),
                                                  false,
                                                  null);

            var forkOp = this;
            //            this._statusWindow.okToClose = false;
            this._statusWindow.addEventListener("load", function() {
                forkOp._statusWindow.enableDoneButton = function() {
                    forkOp._statusWindow.document.getElementById("doneButton").style.visibility = "visible";
                }
            }, true);
//             this._statusWindow.setTimeout( function() {
//                 if ( forkOp._statusWindow.okToClose ) {
//                     forkOp._statusWindow.close();
//                     forkOp._statusWindow = null;
//                 } else {
//                     forkOp._statusWindow.okToClose = true;
//                 }
//             }, 5000);
            syncToMain();
            return;
        }

        // Display the windows side-by-side to help the comparison
        var mainDOM = this.getMainDOMWindow();
        var forkDOM = this.getDOMWindow();

        var continuationFn = function(aSyncToMain) {
            forkOp.killForkFocus();

            if (aSyncToMain) {
                syncToMain();
            } else {
                // If we don't do a load in the main window, we at least need to stop mirroring
                // and record the decision
                if ( aSyncToMain != null ) { // interpret null as "ignore"/"no decision"                
                    forkGlobalState.resetVisitCount(tailDomain);
                    forkOp.stopMirroring("Difference detected; main path chosen");
                    forkOp.recordNegativePermissions(url);
                    // forkOp.clearAcceptedCookies();
                } else {
                    //                forkOp.stopMirroring("Difference detected, but ignored by user (no decision)");
                    forkOp._doSync = true;
                }
            }

            mainDOM.moveTo(oldLeft, oldTop);
            mainDOM.resizeTo(oldWidth, oldHeight);

            if ( oCookieLog.debug ) {
                forkDOM.resizeTo(800,600);
            } else {
                hideDOMOrChromeWindow(forkDOM);
            }

            var doc = forkOp.getDOMWindow().document;
            var elt;
            if ( elt = doc.getElementById("diffcallout") ) {
                elt.parentNode.removeChild(elt);
            }
            if ( elt = doc.getElementById("diffhilite") ) {
                elt.parentNode.removeChild(elt);
            }
        }
        
        // First, arrange the windows, then get the user's answer, then put the main window back afterwards
        var oldLeft = mainDOM.screenX;
        var oldTop = mainDOM.screenY;
        var oldWidth = mainDOM.outerWidth;
        var oldHeight = mainDOM.outerHeight;

        var width = mainDOM.screen.availWidth;
        var height = mainDOM.screen.availHeight;
        
        dumpToConsole("screen width=" + width + ", screen height=" + height);

        var dlgWidth = 420;
        var dlgHeight = 620;

        var newOuterWidth = Math.max ( ((width - dlgWidth) / 2), 500 );           
        
        mainDOM.moveTo(0, 0);        
        mainDOM.outerWidth = newOuterWidth;
        mainDOM.outerHeight = height;
        
        forkDOM.outerWidth = newOuterWidth;
        forkDOM.outerHeight = height;
        //            alert("moved fork to " + ( width - newOuterWidth ) );
        forkDOM.moveTo( width - newOuterWidth, 0 );
        
        this.noKillForkFocus();
        forkDOM.focus(); //bring to front
        mainDOM.focus(); //bring to front       

        compareWin = window.openDialog("chrome://doppelganger/content/comparison.xul", 
                                       "comparison",
                                       "chrome,dialog=no,alwaysRaised=yes"
                                       + ",left=" + ( (width/2) - (dlgWidth/2) + 12) 
                                       + ",top=" + ( (height/2) - (dlgHeight/2) )
                                       + ",outerWidth=" + dlgWidth + ",outerHeight=" + dlgHeight
                                       ,
                                       function (arg) { continuationFn(arg)  },
                                       this.cookiePermissions,
                                       url,
                                       this._lastP3PHeader
                                       );
    };

    this.notifyRequest = function(oHttp, url, method) {
        if ( ! this.stillMirroring() ) {
            oHttp.cancel(NS_BINDING_ABORTED);
            return;
        }

        if ( method.toUpperCase() == "POST" ) {
            dumpToConsole("POST in fork window; not mirroring", "forkwindow-debug");
            oHttp.cancel(NS_BINDING_ABORTED);
            this.stopMirroring("POST in fork window; not mirroring");
        }
    };

    // This is how we make sure that the two windows look the same
    // We hear the STARTs and STOPs and when both window are done we compare them
    this.setTopLevelRequest = function(requestURL, DOMWindow) {
        var DOMWindowUnwrapped = DOMWindow.wrappedJSObject ? DOMWindow.wrappedJSObject : DOMWindow;
        DOMWindowUnwrapped.topLevelRequest = requestURL;
        DOMWindowUnwrapped.originalTopLevelRequest = requestURL;        
    }

    this.visitNewSite = function(tailDomain, url) {
        this.stepCount = 1;
        dumpToConsole("Resetting stepCount to 1 in visitNewSite", "forkwindow-debug");
        this._doSync = false;
        this._lastTailDomain = tailDomain;
        this._tailDomainRE = new RegExp(this._lastTailDomain + "$", "i");
        this.lastClickIdx = null;


        //        this.clearAcceptedCookies();
        

        // Before loading the new page, make sure that there is anything to do, 
        // i.e., that the fork window will accept more cookies than the main page
        // and that decisions that it would make haven't already been made
        var intMainPerms = cookieperms.getPermissions( this.getMainDOMWindow(), 
                                                       tailDomain );
        
        var useful = false;

        var alreadyDidLogin = ( forkGlobalState.doneDomains[tailDomain] != null );

        if ( ! alreadyDidLogin && ! wasDecision(intMainPerms, FP_SESSION) ) {

            if ( this.cookielog.paranoia == "paranoia-low" ) {
                dumpToConsole("Enabling FP_SESSION for " + tailDomain + " because of low paranoia setting", "forkwindow-debug");
                cookieperms.setDomainCookiePerms(tailDomain, FP_SESSION);
                forkGlobalState.doneDomains[tailDomain] = 1;
            } else {
                this.cookiePermissions = FP_SESSION;
                dumpToConsole("Fork window enabling FP session cookies for site " + tailDomain);
                useful = true;
            }

        } else if ( ! alreadyDidLogin && ! wasDecision(intMainPerms, FP_PERSISTENT) ) {
            this.cookiePermissions = FP_SESSION | FP_PERSISTENT;
            dumpToConsole("Fork window enabling FP persistent cookies for site " + tailDomain);
            useful = true;
        }
        
        if (useful) {
            var visitStatus = forkGlobalState.visitSite(tailDomain);
            if ( visitStatus == 0 ) {
                dumpToConsole("Visit count exceeded for " + tailDomain + "; recording negative permissions");
                this.recordNegativePermissions(url);
                this.stopMirroring("Not mirroring " + tailDomain + " because decision already made");
                return;
            } else if ( visitStatus > 0 ) {
                dumpToConsole("Fork window not mirroring " + tailDomain  
                              + " because no new decisions would be made.", "forkwindow-debug");
                this.stopMirroring("Not mirroring " + tailDomain + " because decision already made due to visit count");
            }

            this._forkDone = this._mainDone = false;
            this.load(url);
        } else {
            dumpToConsole("Fork window not mirroring " + tailDomain  
                          + " because no new decisions would be made.", "forkwindow-debug");
            this.stopMirroring("Not mirroring " + tailDomain + " because no new decisions would be made");
        }
    };

    //
    // p3pHeader is sent only for the fork window and only with STOPs; it
    // may be null if none was found in the HTTP response
    //
    //
    // FIXME: One possible bug can happen if the fork window STARTs and
    // STOPs a load before the main window even STARTs. It's unlikely, but
    // it can happen, because the mirrored click action happens first. The
    // opposite situation can't happen, because a main window START results
    // in a load in the fork window right away. To really solve this, we'd
    // need to keep track of epoch numbers or something, but I am trying to
    // avoid that since this function is too complex already.
    //
    this._waitingFor = null; // "fork" or "main"
    this._waitingTimer = null; // a timeout handle to detect loads unique to one page or the other
    this.notifyStateChange = function(isMain, isStart, status, requestURL, p3pHeader, isHiddenRedirect) {
        dumpToConsole("win=" + (isMain ? "MAIN" : "FORK") + 
                      " isStart=" + (isStart ? "START" : "STOP") +
                      " status=" + status + " requestURL=" + requestURL, "statechange-debug");

        var thisLocal = this;
        var waitTimer = function() {
            thisLocal._waitingTimer = null;
            thisLocal._waitingFor = null;
            dumpToConsole("Wait timer fired", "forkwindow-debug");
            if ( thisLocal.stillMirroring() && ! thisLocal.compareAgainstMain()  ) {
                dumpToConsole("Waited too long for a load and pages are different; suspecting an extra load",
                              "forkwindow-debug");
                thisLocal.suspectDifference();
            }
        };

        if ( isStart ) {

            if( ! requestURL ) {
                return;
            } else {
                url = canonicalizeURL(requestURL);
            }

            if ( ! url.match(/^http/) ) {
                this.stopMirroring("Cannot mirror non-HTTP sites");
                return;
            }
 
 
            var tailDomain = getTailURLDomain(getURLDomain(url));


            if ( isMain ) {

                // Prevent the main window from getting one load ahead of the fork window
                // Usually caused by a click by the user before the fork window is done loading:
                // START(main)-START(fork)-STOP(main)-START(main)[new page]
                if ( this.stillMirroring() ) {
                    if ( this._waitingFor == "fork" && ! isHiddenRedirect ) {
                        dumpToConsole("Second load in main detected while previous load not done in fork window; re-syncing",
                                      "forkwindow-debug");
                        this.load(requestURL);
                    } else  if ( ! this._mainDone && ! isHiddenRedirect ) {
                        // A weird, irritating, non-matching-start-and-stop case:
                        // START(main)-START(fork)-START(main)[new page]
                        // Happens if user clicks a link on a page quickly after another one, while the 
                        // first is still loading. Can also happen with redirects, so we filter those out
                        dumpToConsole("A second load in main detected while first load active; probably a quick second click; re-syncing",
                                      "forkwindow-debug");
                        this.load(requestURL);
                    }
                }

                this._mainDone = false;
                if ( ! this._waitingFor ) {
                    this._waitingFor = "fork";
                    this._waitingTimer = setTimeout( waitTimer, 3000 );      
                    dumpToConsole("Waiting for fork to start loading: started main", "forkwindow-debug");
                } else if ( this._waitingFor == "main" ) {
                    dumpToConsole("Not waiting for main any more", "forkwindow-debug");
                    clearTimeout(this._waitingTimer);
                    this._waitingTimer = null;
                }
                
                if ( (tailDomain != this._lastTailDomain) ||
                     ( this.cookiePermissions == FP_SESSION && this.cookielog.paranoia == "paranoia-low") ) {
                    this.visitNewSite(tailDomain, url);
                    return;
                } else { // not a new site
                    if ( ! this.stillMirroring() ) { 
                        this.setSync("NOFORK");
                        return;
                    } else if ( this._doSync ) {
                        this._doSync = false;
                        this.load(requestURL);                        
                    }

                    if ( this._clickSinceLastStart ) {
                        this.stepCount++;
                        this._clickSinceLastStart = false;
                    }

                    dumpToConsole("Increasing stepCount to " + this.stepCount, "forkwindow-debug");

                    // Check if this step put us over the limit;
                    // if so, assume that the cookies aren't needed
                    if ( ! this.stillMirroring() ) {
                        this.setSync("NOFORK");
                        this.recordNegativePermissions(url);
                        dumpToConsole("Step count exceeded max (" + forkGlobalState.SITE_MAX_STEP_COUNT + "); recording negative perm",
                                      "forkwindow-debug");
                        this.overlayStatusMessage("Mirroring stopped because stepCount exceeded " + forkGlobalState.SITE_MAX_STEP_COUNT);
                    }
                }
            } else { // START in fork window
                if ( (! this._loadPending) && 
                     (tailDomain != this._lastTailDomain) ) {
                    this.stopMirroring("Fork window cannot leave domain on its own");                  
                    return;
                }
                
                //                dumpToConsole("waiting for " + this._waitingFor);
                this._forkDone = false;
                if ( ! this._waitingFor && ! this._loadPending ) {
                    this._waitingFor = "main";
                    this._waitingTimer = setTimeout( waitTimer, 3000 );
                    dumpToConsole("Waiting for main to start loading; started fork", "forkwindow-debug");
                } else if ( this._waitingFor == "fork" ) {
                    dumpToConsole("Not waiting for fork any more", "forkwindow-debug");
                    clearTimeout(this._waitingTimer);
                    this._waitingTimer = null;
                }

                this.setSync("LOADING");
                this._loadPending = false;
            }
            
            return;
        }

        // (else) STOP Event

        if ( this._statusWindow ) {
            if ( this._statusWindow.closed ) {
                this._statusWindow = null;
            } else {
                this._statusWindow.enableDoneButton();
            }
        }

        // A load causes the old request to stop, before the new START, so ignore it
        if ( this._loadPending && isMain == false && this._forkDone == true ) {
            return; // the load made the old request cancel
        }
        
        if ( isMain ) {
            this._mainDone = true;
            if ( this._waitingFor == "main" ) {
                dumpToConsole("Not waiting for " + this._waitingFor + " any more", "forkwindow-debug");
                clearTimeout(this._waitingTimer);
                this._waitingFor = null;
            }
        } else {
            this._forkDone = true;
            if ( this._waitingFor == "fork" ) {
                dumpToConsole("Not waiting for " + this._waitingFor + "any more");
                clearTimeout(this._waitingTimer);
                this._waitingFor = null;
            }
        }


        // If there was an error, then just re-sync at the next opportunity
        if ( status != NS_OK ) {

            dumpToConsole("Error loading " + (isMain ? "main" : "fork") + " window; status = " + status, "forkwindow-debug");

            //            this._doSync = (isMain ? "main" : "fork");
            if ( status == 2152398850 && ! isMain && this.stillMirroring() ) {
                this.setSync("NOSYNC");
                //                this.overlayStatusMessage("Load canceled in fork window (maybe a POST)");
            } else if ( this.stillMirroring() ) {
                this.setSync("NOSYNC");
                //                this.overlayStatusMessage("Error loading " + this._doSync + " window; status = " + status);
            }
            return;
        } 
        // else { 
        //             // we can get an error code and then *another* STOP without an
        //             // error if a page messes with window.location
        //             var thisWin = (isMain ? "main" : "fork");
        //             if ( this._doSync == thisWin ) {
        //                 this.overlayStatusMessage("");
        //                 this._doSync = false;
        //             }
        //         }

        if ( this._doSync ) {
            return;
        }
        
        if ( this.stillMirroring() ) {
            this.overlayStatusMessage("");
        }

        if ( isMain ) { // main STOP
            // We need to wait until both fork and main are done loading
            // before making any comparisons
            if ( this._forkDone && this._waitingFor != "fork" ) {
                this.setSync("SYNC");
                if ( this.stillMirroring() &&
                     ! this.compareAgainstMain() ) {
                    this.suspectDifference();
                }
            }
            
        } else { // fork STOP
            if ( this._waitingFor != "fork" ) {
                this.setSync("SYNC");
            }
            this._lastP3PHeader = p3pHeader;

            // We need to wait until both fork and main are done loading
            // before making any comparisons
            if ( this._mainDone ) {
                if ( ! this.compareAgainstMain() &&
                     this.stillMirroring() ) {
                    this.suspectDifference();
                }
            }
        }
    };

    this.load = function(url) {
        this._forkDone = false;
        if ( this._waitingFor == "main" ) {
            this._waitingFor = null;
        }
        this._loadPending = true;

        dumpToConsole("Issued a load in fork window for " + url, "forkwindow-debug");

        var domWin = this.getDOMWindow();
        if ( ! domWin ) {
            return false;
        }
        domWin.location = url;
        //        notifyObservers(null, "load-into-fork", url);
        return true;
    };

    this.observe = function(frm, aTopic, tailDomain) {
        if ( aTopic == "login-form-detected" ) {
            var domainPerms = cookieperms.getDomainCookiePerms(tailDomain);
            if ( ( domainPerms & FP_SESSION ) == 0 ) {

                domainPerms = FP_SESSION;
                cookieperms.setDomainCookiePerms(tailDomain, domainPerms);

                forkGlobalState.doneDomains[tailDomain] = 1;
                forkGlobalState.resetVisitCount(tailDomain);
                
                dumpToConsole("Enabling FP session cookies for " +
                              tailDomain +
                              " because a login form was detected");
                this.copyCookiesToMain(false, tailDomain);
                this.clearAcceptedCookies(false, tailDomain);

            };
        }
    };
    
    // msg can contain HTML
    this.overlayStatusMessage = function(msg) {
        var forkDOM = this.getDOMWindow();
        if ( ! forkDOM ) {
            return;
        }

        var overlayWin = ( this._isFrameset(forkDOM) ) ? forkDOM.frames[0] : forkDOM; 

        var doc = overlayWin.document;

        var OVERLAY_MSG_ID = "CookieForkMsg";
        
        if ( ! doc ) {
            return;
        }

        try {
            var elt;
            if ( elt = doc.getElementById(OVERLAY_MSG_ID) ) {
                elt.parentNode.removeChild(elt);
            }
        }
        catch(e) {
            // if it's not there (or whatever), who cares
        }
        
        if ( ! msg ) {
            return;
        }

        //        doc.body.style.minWidth = 500;

        var overlay = doc.createElement("div");
        overlay.setAttribute( "style", 
                              "background-color: #CCFFCC; " +
                              "display: table; border-collapse: collapse; " +
                              "font-family: Arial, sans-serif; " +
                              "font-size: 18pt;" +
                              "border: solid black 2px; " +
                              "width: 50%; height: 20%; " +
                              "left: 25%; top: 40%; " +
                              "z-index: 999; " +
                              "text-align: center; " +
                              "position: absolute; ");
        overlay.id = OVERLAY_MSG_ID;
        overlay.innerHTML = "<large style='display: table-cell; margin: 4em; text-align: center; vertical-align: middle; position: relative; top:50%'>" +
        msg + "</large>";

        try {
            if ( doc.body ) {
                doc.body.appendChild(overlay);
            } else {
                doc.appendChild(overlay);
            }
        } catch (e) {
            // FIXME: try harder?
        }
    };

    return this;
};
