// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****
//


// 
// This file implements the ReplayState object, which manages the list
// of things to replay, and gets callbacks during the replay which it
// uses to advance its internal state and initiate further replay
// triggers. It is instantiated as the replayState member of
// oCookieLog in the cookielog.cookieReplay function, which currently
// implements the Fixme button.
//
// This file is intended to live in the same scope as CookieLog.js
// and CookiePermissions.js

//
// The three arguments are the cookielog object that spawned this, the
// target window into which we are replaying, and an initial list of
// states (START, STOP, and TRIGGER rows from the log) to replay.
// 
function ReplayState(cookielog, win, replayList) {
    this.cookielog = cookielog;
    this.START = cookielog.START;
    this.TRIGGER = cookielog.TRIGGER;
    this.STOP = cookielog.STOP;

    // FIXME: For now, we have not factored out trigger form data from
    // the main tree, so the form data row must follow the
    // corresponding trigger row. That is, we don't save the form data
    // in the replay list.
      
    this.targetWin =  win;    // where the events are going (XUL window container)
    this.states = replayList;	// triggers and starts we will replay/expect
    this.currentIndex = 0;	// index into states array
    this.loadingURI = null;   // which page is currently loading due to a trigger

    this.STARTWAITTIME = 3000; // ms to wait for a start when one is expected
    this.TRIGGERWAITTIME = 500; // ms to wait for a start when none is expected
    this.waitingForStart = null; // if expecting a start, the id from setTimeout
    this.waitingForNoStart = null; // if not expecting a start, the id from setTimeout
    this.doingConfirm = false;

    this._tailDomainRE = null; // which domain are we replaying on

    for (var i = 0; i < this.states.length; i++) {
        if ( this.states[i].type == this.START ) {
            this._tailDomainRE = new RegExp(getTailURLDomain(getURLDomain(this.cookielog.parseStart(this.states[i].data))) + "$", "i");
            break;
        }
    }


    // Did the replay do something unexpected, e.g. START when not
    // expected or no START when it was expected, showElement could not
    // find anything to click on, etc.
    this.unexpected = false;

    this.sanityCheck = function() {
        var lastStopURL = null;
        var betweenStartAndStop = false;
        var anyStartsYet = false;

        if ( this.states.length < 1 ) {
            return true;
        }

//         dumpToConsole("Pre-Sanity Check replay has following states:", "replay-debug");
//         for (var i = 0; i < this.states.length; i++) {
//             dumpToConsole("\t (" + i + ")" + this.states[i].data, "replay-debug");
//         }

        if ( this._tailDomainRE == null ) {
            dumpToConsole("Internal Error: Could not determine tailDomain for replay", "replay-debug");
            this.cookielog.stopReplaying();
            return false;
        }

        // First do cleanup operations

        // Check for two starts in a row, kill the initial STARTs
        for (var i = 0; i < this.states.length; i++) {
            while ( this.states[i].type == this.START &&
                    i < this.states.length - 1 &&
                    this.states[i + 1].type == this.START ) {

                        //                   dumpToConsole("Internal Error: Replay had two starts in a row: "
                        //                                 + this.states[i].data + " " + this.states[i + 1].data);
                        //                  this.cookielog.stopReplaying();
                        //                  return false;
                  
                        // Filter out the first START: it's usually a load that was implicitly
                        // canceled by a second click on a different link
                        this.states.splice(i, 1);
                    }
        }
 
//         // FIXME: This can happen if e.g. Javascript replaces window.location
//         // Make sure that every START has a trigger for it (no STOP - START sequences)
//         for ( var i = this.states.length - 1; i >= 1; --i ) {
//             if ( this.states[i].type == this.START &&
//                  this.states[i-1].type == this.STOP ) {            
//                      dumpToConsole("Internal Error: START without preceding trigger:");
//                      dumpToConsole("\t" + this.states[i].data);
//                  }
//         }
           
        // Check that there are no triggers between STARTs and STOPs;
        // Fix this by moving STOPs right after the previous START
        for ( var i = this.states.length - 1; i >= 1; --i ) {
            if ( this.states[i].type == this.STOP &&
                 this.states[i-1].type != this.START ) {
                       
                     stopEntry = (this.states.splice(i,1))[0];
                       
                     for ( var j = i-1; j >= 0; --j ) {
                         if ( this.states[j].type == this.STOP ) {
                             dumpToConsole("Internal Error: Two STOPs in a row:\n"
                                           + "\t" + this.states[j].data + "\n\t" + stopEntry);
                             return false;                     
                         }

                         if ( this.states[j].type == this.START ) {
                             this.states.splice(j + 1, 0, stopEntry);
                             break;
                         }

                     }

                     i = j + 1; // skip to where we inserted the STOP
                 }
        }

//         // FIXME: For now, kill all chrome elements
//         for (var i = 0; i < this.states.length; i++) {
//             if ( this.states[i].type == this.TRIGGER &&
//                  this.states[i].data.match("chrome://" ) ) {
//                      this.states.splice(i, 1);
//                      --i; // re-process the same index since we removed the element
//                  }
//         }


        // trailing chrome trigger is the FIXME button; kill it
        if ( this.states[this.states.length - 1].type == this.TRIGGER &&
             this.states[this.states.length - 1].data.match("chrome://") ) {
                 this.states.splice(this.states.length - 1, 1);
             }


        // Now check that everything is OK
        for (var i = 0; i < this.states.length; i++) {

            if ( this.states[i].data.indexOf("chrome:") >= 0 ) {
                continue;
            }
              
            if ( this.states[i].type == this.START ) {
                betweenStartAndStop = true;
                anyStartsYet = true;
            }

            if ( this.states[i].type == this.STOP ) {
                lastStopURL = (this.cookielog.parseStop(this.states[i].data))[0];
            }

            if ( ! anyStartsYet )
            continue;

            // REMOVED this check because redirects break it and we don't check URLs anyway
            //              // Make sure starts and stops correspond
            //               if ( this.states[i].type == this.STOP ) {
            //                   var stopData = this.cookielog.parseStop(this.states[i].data);
            //                   if ( stopData[0] != lastStartURL ) {
            //                       dumpToConsole("Internal Error: STOP URL differs from previous START: "
            //                                     + this.states[i].data + " " + lastStartURL);
            //                       this.cookielog.stopReplaying();
            //                       return false;                     
            //                   }

            //                   betweenStartAndStop = false;
            //               }
              
            // 	  check that each trigger URL is the same as the last STOP;
            // quick-n-dirty to match either enclosing page or main page for IFRAMEs
            if ( this.states[i].type == this.TRIGGER &&
                 this.states[i].data.indexOf(lastStopURL) < 0 ) {
                internalError("Replay trigger on different page from previous STOP:\n "
                              + this.states[i].data + "\n" + lastStopURL);
                return false;
            }
        }

        return true;
    };

    this.go = function() {
        dumpToConsole("======================================", "replay-debug");
        dumpToConsole("Starting Replay with following states:", "replay-debug");
        for (var i = 0; i < this.states.length; i++) {
            dumpToConsole("\t (" + i + ")" + this.states[i].data);
        }
        dumpToConsole("======================================", "replay-debug");

        if ( this.states.length < 1 ) {
            this.cookielog.stopReplaying();
            return;
        }


        //
        // Copy cookies from fork to main (without overwriting existing ones)
        //
        var cookieManager = getCookieManager();
        var mainCookieList = new Array();
        var cookies = cookieManager.enumerator;
        while (cookies.hasMoreElements()) {
            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie2);
            mainCookieList[cookie.host+cookie.name+cookie.path] = 1;
        }

        var prefix = "FORK";
        cookies = cookieManager.enumerator;
        while (cookies.hasMoreElements()) {

            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie2);

            // FIXME: For now assume that replay is FP_SESSION | FP_PERSISTENT
            // FIXME: Also assume the replay is on one domain only

            if ( cookie.host.match(this._tailDomainRE) &&  
                 cookie.name.substr(0, prefix.length) == prefix ) 
                {
                    var newCookie = new Cookie(cookie);
                    newCookie.name = newCookie.name.substr(prefix.length);
                    
                    if ( ! mainCookieList[newCookie.host+newCookie.name+newCookie.path] ) {
                        addCookieObject(newCookie);
                    }                    
                    
                    cookieManager.remove(cookie.host, cookie.name, cookie.path, false);
                }

        }

        
        //
        // Now launch the replay itself
        //

        this.currentIndex = 0;
        this.doingInitialLoad = true; 

        if ( this.states[0].type == this.TRIGGER ) {            

            // Just load the initial page; there may be many clicks before the next load

            this.cookielog.showElement(this.targetWin, this.states[0].index, true, false, false);

        } else if (this.states[0].type == this.START ) {

            htmlWinFromXULWindow(this.targetWin).location = this.cookielog.parseStart(this.states[0].data);

        } else {

            internalError("Don't know how to start a replay with other than START or TRIGGER");            
            this.cookielog.stopReplaying();

        }
    }
       
    this.isDoneReplaying = function() { return this.currentIndex >= this.states.length; };

    this.setWaitStartTimer = function() {
        this.waitingForStart = setTimeout(this.startDidntHappen.bind(this), this.STARTWAITTIME);
    };

    this.setWaitNoStartTimer = function() {
        this.waitingForNoStart = setTimeout(this.next.bind(this), this.TRIGGERWAITTIME);
    }

    this.startDidntHappen = function () { 
        this.unexpected = true;
        var errmsg = "Expected page load did not occur in time; stopping replay.";
        //        this.targetWin.alert(errmsg);
        dumpToConsole(errmsg, "replay-debug");
        this.cookielog.stopReplaying();
    };


    // Do the "next thing" in the list: send a trigger, then wait appropriately
    this.next = function() {
        if ( this.waitingForNoStart ) {
            dumpToConsole("As expected, no start occurred. Going to next (trigger) state.", "replay-debug");
        }
        this.waitingForNoStart = null;

        if ( this.unexpected ) {
            return;
        }

        if ( this.isDoneReplaying() ) {
            throw "Can't continue replay: reached end of list!";
        }
	  
        if ( ! this.advance() )
        {
            // done!
            cookielog.stopReplaying();             
            this.targetWin.alert("Replay completed successfully!\nHopefully your problem is solved.");
            return;
        }

        switch (this.states[this.currentIndex].type) {
            case this.TRIGGER: 
            // wait a while before next trigger, to make sure
            // no STARTs come first
            var nextIsStart = false;
            if (this.currentIndex < this.states.length - 1) {
                if ( this.states[this.currentIndex + 1].type != this.START ) {
                    this.setWaitNoStartTimer();
                } else {
                    nextIsStart = true;
                    this.setWaitStartTimer();
                }
            }                

            if ( ! this.playTrigger() ) {
                // Cancel timers if something went wrong
                this.waitingForStart = this.waitingForNoStart = null;
                //            this.targetWin.alert("Could not play trigger: canceling timers and stopping replay\nPerhaps the loaded page was different?");
                    this.cookielog.stopReplaying();
                return;
            }
            
            if ( nextIsStart) {
                this.next();
            }
            // Else waitnostart will call next()
            break;

            case this.START:
            // The timer that checks for this is set in the trigger
            // handler, because it needs come before the trigger is fired
            dumpToConsole("Passed through START state in next()", "replay-debug");
            break;

            case this.STOP:
            // FIXME: Have some sort of load timeout (or possible rely on error notifications)
            // For now, this.stoppedLoading calls next() again
            dumpToConsole("Passed through STOP state in next()", "replay-debug");
            break;

            default:
            internalError("Unknown replay list entry type: " +
                          this.states[this.currentIndex + 1].type);
        }
    };

    // throws exception if current state is invalid
    //
    // Should pretty much only be called from next() since otherwise the
    // resulting START may be "unexpected"
    this.playTrigger = function() {
        try {
            if ( this.unexpected ) {
                var msg = "Unexpected event already terminated replay in playTrigger";
                internalError(msg);
                throw new Error("INTERNAL ERROR: " + msg);
            }

            if ( this.isDoneReplaying() ) {
                throw new Error("Can't replay trigger: reached end of list!");
            }

            if ( this.states[this.currentIndex].type != this.TRIGGER ) {
                var msg = "ReplayState.playTrigger called while not at TRIGGER state";
                internalError(msg);
                throw new Error("INTERNAL ERROR: " + msg);
            }
            
            var clickString = this.states[this.currentIndex].data;

            dumpToConsole("Replaying trigger in state " + this.currentIndex + ": " + clickString, "replay-debug");
            
            if ( clickString.match("chrome://") ) {
                var label = isAllowedChromeButton(clickString);
                if ( label ) {
                    chromeButtonLabelToAction(this.targetWin, label);
                }
            } else {
                this.cookielog.showElement(this.targetWin, 
                                           this.states[this.currentIndex].index, 
                                           false, true, false);
            }
        }
        catch (err) {
            this.cookielog.stopReplaying();
            this.targetWin.alert("Replay failed: " + err);
            dumpToConsole("Replay failed: " + err, "replay-debug");
            return false;
        }

        return true;
    };
	  
    // go to next state in the list; returns false if nothing left
    // returns true if everything is ok; throws exception on error
    this.advance = function() {
        if ( this.unexpected ) {
            internalError("Unexpected event already terminated replay in playTrigger");
            throw new Error("INTERNAL ERROR: Unexpected event already terminated replay in playTrigger");
        }

        if (this.currentIndex >= this.states.length) {
            debugger;
            return false;
        }
	  
        ++this.currentIndex;

        dumpToConsole("[REPLAY] Advancing to state " + this.currentIndex, "replay-debug");
	  
        if (this.currentIndex >= this.states.length) {
            return false;
        }
	  
        return true;
    };

    // called when a START event is received during replay
    this.startedLoading = function(url) {
        dumpToConsole("[REPLAY] Got a START during replay: " + url, "replay-debug");

        if ( this.doingInitialLoad  || this.unexpected ) {
            if ( this.states[this.currentIndex].type == this.START ) {
                this.next();
            }
            return;
        }
          
        this.loadingURI = url;
          
        if ( this.waitingForStart == null ) {
            this.unexpected = true;
            var errmsg = "A page loaded during the replay when one didn't before. Hopefully the problem is solved.";
            //            this.targetWin.alert(errmsg);
            dumpToConsole(errmsg, "replay-debug");
            this.cookielog.stopReplaying();
            return;
        } else {
            clearTimeout(this.waitingForStart);
            this.waitingForStart = null;
            this.next();
//             if ( ! ( this.advance() ) { // go past START to STOP
//                 alert("Replay completed successfully!\nHopefully your problem is solved.");
//                 this.cookielog.stopReplaying();
//                 return;
//             }
        }
    };

    // indicates that a replay-induced page finished loading
    // FIXME: look at error codes to see if the load worked
    this.stoppedLoading = function(domWindow) {
        dumpToConsole("[REPLAY] Got a STOP during replay: " + domWindow.location.href, "replay-debug");
        //          alert("Got STOP: " + domWindow.document.title);

        if ( this.doingInitialLoad ) {
            this.doingInitialLoad = false;
            this.next();         // and now we're off to the races
            return;
        }

        if ( this.states[this.currentIndex].type != this.STOP ) {
            this.unexpected = true;
            var errmsg = "Replay: got STOP event without expecting one; stopping replay.";
            dumpToConsole(errmsg, "replay-debug");
            dumpToConsole("Expecting " + this.states[this.currentIndex].data, "replay-debug");
            //            this.targetWin.alert(errmsg);
            this.cookielog.stopReplaying();
            return;
        }

        var newTitle = domWindow.document.title;	  
        var stopData = this.cookielog.parseStop(this.states[this.currentIndex].data);
          
        //           dumpToConsole("In STOP, expected title/got title: ");
        //           dumpToConsole("\t" + stopData[1]);
        //           dumpToConsole("\t" + newTitle);

        if ( newTitle != stopData[1] ) {
            this.unexpected = true;
            var errmsg = "A different page loaded than before; perhaps the cookie setting changed it.\nStopping replay.\n";;
            //            this.targetWin.alert(errmsg);
            dumpToConsole(errmsg, "replay-debug");
            dumpToConsole("\told title = " + stopData[1], "replay-debug");
            dumpToConsole("\tnew title = " + newTitle,    "replay-debug");
            this.cookielog.stopReplaying();
            return;
        }

        this.loadingURI = null;
        this.next(false);
    };

    this.madePostRequest = function(url, oHttp) {
        var wfs = false;
        var wfns = false;
        if ( this.waitingForStart ) {
            clearTimeout(this.waitingForStart);
            this.waitingForStart = null;
            wfs = true;
        } else if ( this.waitingForNoStart ) {
            clearTimeout(this.waitingForNoStart);
            this.waitingForNoStart = null;
            wfns = true;
        }

        var proceed = this.targetWin.confirm ("Doppelganger is about to replay what may be a transaction.\n" + 
                                              "If this is a purchase or other action you don't want to replay,\n" +
                                              "press Cancel now. Otherwise, press OK" ); 
        
        if ( proceed ) {
            if ( wfs ) { this.setWaitStartTimer(); }
            else if ( wfns ) { this.setWaitNoStartTimer(); }
            
            return;
        }

        oHttp.cancel(NS_BINDING_ABORTED);
        this.unexpected = true;
        var errmsg = "A POST request was encountered while replaying (user canceled). Stopping replay.";
        //this.targetWin.alert(errmsg);
        dumpToConsole(errmsg, "replay-debug");        
        this.cookielog.stopReplaying();
    }

    // FIXME: There is a bug in JS: https://bugzilla.mozilla.org/show_bug.cgi?id=300079
    // Total hack until I work out the prototype problem here (this.method.constructor != Function)
    addBindTo(this.next);
    addBindTo(this.startDidntHappen);
};

